﻿// adds getter and setter methods to all properties on an object
var addGetSet = function (obj) {
    for (var n in obj) {
        var newKey = '_' + n;
        var objType = Lava.typeOf(obj[n]);
        if (obj.hasOwnProperty(n) && !Object.getOwnPropertyDescriptor(obj, n).get) {
            if (objType === 'object') {
                addGetSet(obj[n]);
            }
            else if (n != 'computedProperties') {
                // add binding storage to the object
                if (!obj.bindings) obj.bindings = {};
                if (!obj.bindings[n]) obj.bindings[n] = [];
                if (objType === 'array') {
                    // add array specific binding storage
                    if (!obj[n].bindings) obj[n].bindings = {};

                    if (!obj[n].bindings.add) obj[n].bindings.add = [];
                    if (!obj[n].bindings.addAt) obj[n].bindings.addAt = [];
                    if (!obj[n].bindings.remove) obj[n].bindings.remove = [];
                    if (!obj[n].bindings.removeAt) obj[n].bindings.removeAt = [];
                }

                if (Lava.typeOf(obj[n]) === 'computed')
                    if (!obj[n].context)
                        obj[n].context = obj;

                obj[newKey] = obj[n];
                delete obj[n];
                (function (object, key, name, type) {
                    Object.defineProperty(obj, name, {
                        get: function () {
                            if (this.dependOn && this.dependOn.indexOf(name) === -1) {
                                this.dependOn.push(name);
                            }
                            return this[key];
                        },
                        set: function (x) {
                            this[key] = x;
                            set(obj, name, obj[name]);
                        },
                        configurable: true,
                        enumerable: true
                    });
                    // set up array methods
                    if (type === 'array') {
                        obj[name].add = function (value) {
                            obj[name].push(value);
                            set(obj[name], 'add', value);
                            return value;
                        };
                        obj[name].addAt = function (index, value) {
                            obj[name].splice(index, 0, value);
                            set(obj[name], 'addAt', { index: index, value: value });
                            return value;
                        };
                        obj[name].remove = function (value) {
                            set(obj[name], 'remove', value);
                            obj[name].splice(obj[name].indexOf(value), 1);
                            return value;
                        };
                        obj[name].removeAt = function (index) {
                            var item = obj[name].splice(index, 1);
                            set(obj[name], 'removeAt', index);
                            return item.length > 0 ? item[0] : null;
                        };
                        obj[name].empty = function () {
                            var len = obj[name].length;
                            while (len--) {
                                obj[name].removeAt(len);
                            }
                        };
                        obj[name].fill = function (lavaArray) {
                            obj[name].empty();
                            var len = lavaArray.length;
                            for (var i = 0; i < len; i++) {
                                obj[name].add(lavaArray[i]);
                            }
                        };
                    }
                }(this, newKey, n, objType));

                if (!obj.computedProperties)
                    obj.computedProperties = [];
            }
        }
    }
};

// sets the changed value of an object's property to all of its bound properties
var set = function (object, property, value) {
    var propertyBindings = object.bindings[property];
    var len = propertyBindings.length;
    for (var a = 0; a < len; a++) {
        if (propertyBindings[a].func) {
            propertyBindings[a].func(value);
        }
        else {
            propertyBindings[a](value);
        }
    }
};

Lava.bind = function () {
    while (document.querySelectorAll('[' + Lava.dataBindName + ']').length > 0) {      
        var elements = document.querySelectorAll('[' + Lava.dataBindName + ']');

        for (var a = 0; a < elements.length; a++) {
            var element = elements[a];
            // get the binding expression
            var expressions = JSON.parse(convertObjectStrToJSON(element.getAttribute('' + Lava.dataBindName + '')));

            for (var attribute in expressions) {
                if (expressions.hasOwnProperty(attribute)) {
                    var expression = expressions[attribute];

                    if (Lava.typeOf(expression) === 'object') {
                        // this is a complex binding with multiple different properties in the binding expression
                        // current properties that make up a complex binding: bind, mode, format, args
                        // resolve the bind property
                        var resolved = resolveObjectExpression(expression.bind);
                        bindExpression(resolved.objectName, resolved.object, resolved.property, element, attribute, expression, resolved.propertyIsIndex);
                    }
                    else {
                        var resolved = resolveObjectExpression(expression);
                        bindExpression(resolved.objectName, resolved.object, resolved.property, element, attribute, expression, resolved.propertyIsIndex);
                    }
                }
            }
            element.removeAttribute(Lava.dataBindName);
        }
    }
};

var bindExpression = function (objectName, object, property, element, attribute, expression, propertyIsIndex) {
    var attribute = resolveAttribute(attribute);
    var func = null;

    // handle template attribute
    if (attribute === 'template' && property == null) {
        element.innerHTML = replaceTemplateTags(document.getElementById(objectName).innerHTML);
        return;
    }

    // handle functions as properties that return a value
    var isNormalFunc = false;
    if (property.slice(-2) === '()') {
        property = property.substring(0, property.length - 2);
        isNormalFunc = true;
    }

    // choose the func based on the type of attribute we are dealing with
    if (Lava.typeOf(object[property]) === 'function') {
        if (isNormalFunc) {
            func = function (value) {
                if (value) {
                    var v = resolveFormatExpressionProperty(value.call(object), expression);
                    if (element[attribute] != v) 
                        element[attribute] = v;
                }
            };
        }
        else {
            if (expression.args) {
                // resolve argument expressions
                var args = {};
                for (var arg in expression.args) {
                    if (expression.args.hasOwnProperty(arg)) {
                        var tmp = resolveObjectExpression(expression.args[arg]);
                        if (tmp.property) {
                            if (tmp.object.hasOwnProperty(tmp.property)) {
                                args[arg] = tmp.object[tmp.property];
                            }
                            else {
                                args[arg] = tmp.property;
                            }
                        }
                        else {
                            // there is no property in the resolved expression so just set the arg to the object
                            args[arg] = tmp.object;
                        }
                    }
                }

                element.addEventListener(attribute, function (e) {
                    object[property].call(object, e, args);
                }, false);
            }
            else {
                element.addEventListener(attribute, function (e) {
                    object[property].call(object, e);
                }, false);
            }
        }
    }
    else if (Lava.typeOf(object[property]) === 'computed') {
        // overwrite dependencies if array var is used
        //var path = getPathFromPropertyExpression(expression);
        //object[property].dependencies = object[property].dependencies.replace(/\[i\]/g, path);
        bindComputedProperty(object, property, element, attribute, expression);
    }
    else if (attribute.indexOf('style.') === 0) {
        var subAttribute = attribute.replace('style.', '');

        func = function (value) {
            value = resolveFormatExpressionProperty(value, expression);
            if (element.style[subAttribute] != value)
                element.style[subAttribute] = value;
        };
    }
    else if (attribute === 'foreach') {
        var array = object[property];

        if (Lava.typeOf(array) === 'array') {
            var tagName = element.tagName.toLowerCase();
            // we need to wrap the children in a parent element if there is more than one child
            if (element.children.length > 1) {
                if (tagName == 'ul') {
                    element.innerHTML = '<li>' + element.innerHTML + '</li>';
                }
                else if (tagName == 'table') {
                    element.innerHTML = '<tr>' + element.innerHTML + '</tr>';
                }
                else if (tagName == 'tr') {
                    element.innerHTML = '<td>' + element.innerHTML + '</td>';
                }
                else {
                    element.innerHTML = '<div>' + element.innerHTML + '</div>';
                }
            }

            var template = element.innerHTML;

            if (expression.template) {
                // get the template
                var templateElement = document.getElementById(expression.template);
                // create the correct type of tempElement.  This is necessary because if we wrap, for example,
                // tr elements with a div to test, the tr will be ignored.
                var tempElement = document.createElement(tagName);
                tempElement.innerHTML = templateElement.innerHTML;
                if (tempElement.children.length > 1) {
                    if (tagName == 'ul') {
                        templateElement.innerHTML = '<li>' + templateElement.innerHTML + '</li>';
                    }
                    else if (tagName == 'table') {
                        templateElement.innerHTML = '<tr>' + templateElement.innerHTML + '</tr>';
                    }
                    else if (tagName == 'tr') {
                        templateElement.innerHTML = '<td>' + templateElement.innerHTML + '</td>';
                    }
                    else {
                        templateElement.innerHTML = '<div>' + templateElement.innerHTML + '</div>';
                    }
                }

                template = templateElement.innerHTML;
                // get rid of the temporary element
                delete tempElement;
            }

            element.innerHTML = '';
            var html = '';

            for (var a = 0; a < array.length; a++) {
                html += replaceTemplateTags(template.replace(/\[i\]/g, objectName + '.' + property + '[' + a + ']'));
            }
            html = resolveAllUnboundExpressions(html);
            element.innerHTML = html;

            // function to handle adding item to array
            var add = function (value) {
                var index = object[property].length - 1;
                var html = replaceTemplateTags(template.replace(/\[i\]/g, objectName + '.' + property + '[' + index + ']'));
                html = resolveAllUnboundExpressions(html);
                element.insertAdjacentHTML('beforeend', html);
                /*setTimeout(function () {*/ Lava.bind(); /*}, 10);*/
                // take care of computed properties
                var computedProperties = array.computedProperties || [];
                for (var a = 0; a < computedProperties.length; a++) {
                    rebindComputedProperty(computedProperties[a], element);
                }
            };
            object[property].bindings.add.push({ func: add, attribute: attribute });
            // function to handle adding item to array at index
            var addAt = function (value) {
                var html = replaceTemplateTags(template.replace(/\[i\]/g, objectName + '.' + property + '[' + value.index + ']'));
                html = resolveAllUnboundExpressions(html);
                var d = document.createElement('div');
                d.innerHTML = html;
                var newNode = d.firstElementChild;
                if (element.children.length === 0)
                    element.innerHTML = html;
                else
                    element.insertBefore(newNode, element.children[value.index]);
                d = null;
                /*setTimeout(function () {*/ Lava.bind(); /*}, 10);*/
                // take care of computed properties
                var computedProperties = array.computedProperties || [];
                for (var a = 0; a < computedProperties.length; a++) {
                    rebindComputedProperty(computedProperties[a], element);
                }
            };
            object[property].bindings.addAt.push({ func: addAt, attribute: attribute });
            // function to handle removing item from array
            var remove = function (value) {
                var index = array.indexOf(value);
                if (index > -1) {
                    element.removeChild(element.children[index]);
                }
            };
            object[property].bindings.remove.push({ func: remove, attribute: attribute });
            // function to handle removing item from array at index
            var removeAt = function (value) {
                if (value > -1) {
                    element.removeChild(element.children[value]);
                }
            };
            object[property].bindings.removeAt.push({ func: removeAt, attribute: attribute });
        }
    }
    else if (attribute === 'if') {
        func = function (value) {
            var shown = element.lava_shown;
            if (Lava.typeOf(shown) === 'undefined') {
                // the if block has not been processed yet
                if (value)
                    element.lava_shown = true;
                else
                    hideElement(element);
            }
            else if (shown === true && value === false) {
                hideElement(element);
            }
            else if (shown === false && value === true) {
                showElement(element);
            }
        }
    }
    else if (attribute === 'ifnot') {
        func = function (value) {
            value = !value;
            var shown = element.lava_shown;
            if (Lava.typeOf(shown) === 'undefined') {
                // the if block has not been processed yet
                if (value)
                    element.lava_shown = true;
                else
                    hideElement(element);
            }
            else if (shown === true && value === false) {
                hideElement(element);
            }
            else if (shown === false && value === true) {
                showElement(element);
            }
        }
    }
    else if (attribute === 'class') {

    }
    else {
        func = function (value) {
            value = resolveFormatExpressionProperty(value, expression);
            if (element[attribute] !== value)
                element[attribute] = value;
        };
    }

    // set up two-way binding for value attribute
    if (attribute === 'value' && expression.mode != 'oneWay') {
        var elementType = element.tagName.toLowerCase()
        var typeAttr = element.getAttribute('type');
        if(typeAttr) 
            elementType += ':' + typeAttr.toLowerCase();
        if ('input:checkbox input:radio select'.indexOf(elementType) > -1) {
            element.addEventListener('change', function () {
                setTimeout(function () { object[property] = element.value; }, 10);
            }, false);
        }
        else if ('input:text textarea'.indexOf(elementType) > -1) {
            element.addEventListener('keypress', function () {
                setTimeout(function () { object[property] = element.value; }, 10);
            }, false);
        }
    }

    // set up two-way binding for checked
    if (attribute === 'checked') {
        var elementType = element.tagName.toLowerCase() + ':' + element.getAttribute('type').toLowerCase();
        if ('input:checkbox input:radio'.indexOf(elementType) > -1) {
            element.addEventListener('change', function () {
                setTimeout(function () { object[property] = element.checked; }, 10);
            }, false);
        }
    }

    // add the default binding function if present.  This will be null for array type functions
    if (func) {
        if (object.bindings) {
            object.bindings[property].push({ func: func, attribute: attribute });
        }
        func(object[property]);
    }
};

var bindComputedProperty = function (object, property, element, attribute, expression) {
    var computed = object[property],
        context = computed.context,
        arr = [];
    if (computed.dependencies) {
        arr = computed.dependencies.split(',');
    } else {
        context.dependOn = arr;
        computed.f.call(context);
        delete context.dependOn;
    }
    computed.properties = [];
    for (var a = 0; a < arr.length; a++) {
        var expressionObject = resolveObjectExpression(arr[a].trim(), context);
        expressionObject.object.computedProperties.push({ computed: computed, element: element, attribute: attribute, expression: expression });
        computed.properties.push(expressionObject);
        rebindComputedProperty({ computed: computed, element: element, attribute: attribute, expression: expression });
    }
};

var rebindComputedProperty = function (object, currentElement) {
    // add binding functions for each dependency object
    for (var a = 0; a < object.computed.properties.length; a++) {
        (function (obj, element, attribute, computed, expression) {
            var f = null;
            if (attribute.indexOf('style.') === 0) {
                var subAttribute = attribute.replace('style.', '');

                f = function () {
                    // do not apply the computed property when it is not on
                    if (!computed.on) return;
                    var value = resolveFormatExpressionProperty(computed.f.call(computed.context), expression);

                    if (!value)
                        value = '';

                    element.style[subAttribute] = value;
                };
            }
            else {
                f = function () {
                    // do not apply the computed property when it is not on
                    if (!computed.on) return;
                    var value = resolveFormatExpressionProperty(computed.f.call(computed.context), expression);

                    if (!value)
                        value = '';

                    element[attribute] = value
                };
            }
            if (Lava.typeOf(obj.object) === 'array') {
                for (var b = 0; b < obj.object.length; b++) {
                    addBinding(obj.object[b].bindings[obj.property], f, attribute, element);
                }
            }
            else {
                addBinding(obj.object.bindings[obj.property], f, attribute, element);
            }
            computed.trigger = f;
            f();
        })(object.computed.properties[a], object.element, object.attribute, object.computed, object.expression);
    }
};

var addBinding = function (existingBindings, func, attribute, element) {
    // make sure that this binding does not already exist
    var exists = false;
    for (var c = 0; c < existingBindings.length; c++) {
        if (existingBindings[c].attribute === attribute && element == existingBindings[c].element) {
            exists = true;
            break;
        }
    }
    if (!exists)
        existingBindings.push({ func: func, attribute: attribute, element: element });
};

var resolveFormatExpressionProperty = function (value, expression) {
    if (expression.format)
        value = expression.format.replace('{0}', value);
    else if (expression.formatExclusive && value != null && value.toString().length > 0)
        value = expression.formatExclusive.replace('{0}', value);

    return value;
};