﻿/// COMPUTED PROPERTY
// Sets up a computed property that depends on other properties in the root object or a child object
// dependencies: optional parameter for describing properties that this computed property should depend on.
// It can have special syntax but in simple case its comma delimited list of property names that this computed property 
// should depend on.  Every time one of these dependent properties changes, a change request will be sent to the computed 
// property as well.
// f: the function that makes up the computed property
Lava.Computed = function (dependencies, f, context) {
    var hasDependencies = typeof dependencies === "string",
        args = Array.prototype.slice.call(arguments);
    if (!hasDependencies) {
        args.unshift(null);
    }
    
    var computed = {
        type: 'computed',
        dependencies: args[0],
        f: args[1],
        on: true,
        context: args[2]
    };

    computed.get = function () {
        return computed.f.call(computed.context);
    };

    return computed;
};

/// END COMPUTED PROPERTY