﻿var unSyncableTypes = ['function', 'computed'];
var complexLavaProperties = ['bindings', 'computedProperties'];

var isSyncableType = function (type) {
    if(unSyncableTypes.indexOf(type) == -1) {
        return true;
    }
    return false;
};

var isComplexProperty = function (property) {
    if (complexLavaProperties.indexOf(property) > -1) {
        return true;
    }
    return false;
};

var showElement = function (element) {
    element.lava_shown = true;
    element.style.display = element.lava_display || '';
};

var hideElement = function (element) {
    element.lava_shown = false;
    element.lava_display = element.style.display;
    element.style.display = 'none';
};

// Determines the type of the object passed in
Lava.typeOf = function (a) {
    if (typeof a === 'undefined') {
        return 'undefined';
    }
    if (a === null) {
        return 'null';
    }
    if (a.type === 'computed') {
        return 'computed';
    }
    if (Object.prototype.toString.call(a) === '[object Array]') {
        return 'array';
    }
    if (Object.prototype.toString.call(a) === '[object Object]') {
        return 'object';
    }
    return typeof a;
};

// Merges a source object into a target object
Lava.extend = function (target, source) {
    var _t = target || {};
    var _s = source || {};

    for (var name in _s) {
        if (_s.hasOwnProperty(name)) {
            var t = _t[name];
            var s = _s[name];

            if (_t === s) {
                continue;
            }
            if (typeof t === 'undefined') {
                _t[name] = _s[name];
            }
        }
    }
};

// Syncs all of the objects properties with the properties of the object passed in
Lava.sync = function (target, source) {
    for (var property in target) {
        if (target.hasOwnProperty(property)) {
            var targetPropType = Lava.typeOf(target[property]);
            var sourcePropType = Lava.typeOf(source[property]);

            if (!isSyncableType(targetPropType) || !isSyncableType(sourcePropType)) {
                continue;
            }
            if (isComplexProperty(property) || isComplexProperty(sourcePropType)) {
                continue;
            }

            if (targetPropType === 'object' && sourcePropType === 'object') {
                Lava.sync(target[property], source[property]);
            }
            else {
                if (property[0] !== '_') {
                    // copy property value
                    target[property] = source[property];
                }
            }
        }
    }
};

// Returns an object without Lava specific properties
Lava.simplify = function (complex, simple) {
    var complex = complex || {};
    var simple = simple || {};

    for (var property in complex) {
        if (complex.hasOwnProperty(property)) {
            var complexPropType = Lava.typeOf(complex[property]);

            if (complexPropType == 'function' || isComplexProperty(property)) {
                continue;
            }

            if (complexPropType === 'object') {
                simple[property] = Lava.simplify(complex[property], simple[property]);
            }
            else {
                if (property[0] !== '_') {
                    if (complexPropType == 'computed') {
                        simple[property] = complex[property].f.call(complex[property].context);
                    }
                    else {
                        simple[property] = complex[property];
                    }
                }
            }
        }
    }
    return simple;
};

// Defers execution of a function until the DOM is ready
Lava.ready = function (f) {
    if (Lava.isReady) {
        f();
    }
    else if (document.addEventListener) {
        document.addEventListener('DOMContentLoaded', function () {
            Lava.isReady = true;
            f();
        });
    }
    else {
        if (/in/.test(document.readyState)) {
            setTimeout(function () { Lava.ready(f); }, 9);
        }
        else {
            Lava.isReady = true;
            f.call(this);
        }
    }
    return this;
};

// Load data from the server using an HTTP GET request
Lava.get = function (url, success) {
    var xhr = new XMLHttpRequest();
    xhr.onreadystatechange = function () {
        if (xhr.readyState == 4) {
            if (xhr.status == 200) {
                if (success) success.call(this, xhr.responseText);
            }
        }
    };
    xhr.open('GET', url, true);
    xhr.timeout = 10000;
    xhr.setRequestHeader('Content-Type', 'text');
    xhr.send('{}');
};

// Performs an asynchronous HTTP request
Lava.ajax = function (options, context) {
    var context = context || window;
    var settings = {
        type: 'POST',
        url: '/',
        contentType: 'application/json; charset=utf-8',
        data: {},
        success: function () { },
        async: true,
        timeout: 10000
    };

    Lava.extend(options, settings);

    var xhr = new XMLHttpRequest();
    xhr.onreadystatechange = function () {
        if (xhr.readyState == 4) {
            if (xhr.status == 200) {
                var text = xhr.responseText.replace(/\\\/Date\((-?\d+)\)\\\//g, function (match, p1) {
                    return new Date(parseInt(p1)).toISOString();
                });
                options.success.call(context, JSON.parse(text));
            }
        }
    };
    xhr.open(options.type, options.url, options.async);
    xhr.timeout = options.timeout;
    xhr.setRequestHeader('Content-Type', options.contentType);
    var data = Lava.simplify(options.data);
    xhr.send(JSON.stringify(data));
};

// Converts an object string to JSON
var convertObjectStrToJSON = function (obj) {
    var nonWritable = '"\'',
        strings = [],
        string = '',
        insideQuotes = false,
        oLen = obj.length,
        json = [];

    // replace brackets
    obj = obj.replace(/\[/g, '<').replace(/\]/g, '>');

    for (var a = 0; a < oLen; a++) {
        var c = obj[a];

        // handle strings inside quotes
        if (c === "'") {
            if (!insideQuotes) {
                string = '';
                insideQuotes = true;
                json.push('[' + strings.length + ']');
            }
            else {
                strings.push(string);
                insideQuotes = false;
            }

            continue;
        }

        if (insideQuotes) {
            string += c;
            continue;
        }
        // END handle strings inside quotes

        // write the char to the output stram
        if (nonWritable.indexOf(c) === -1) {
            json.push(c);
        }
    }

    var jsonStr = json.join('');

    // wrap all keys and values with double quotes
    jsonStr = jsonStr.replace(/[^ \{\},:]+/g, function (a) {
        if (!isNaN(parseFloat(a)) && isFinite(a)) return a;
        return '"' + a + '"';
    });

    for (var a = 0; a < strings.length; a++) {
        jsonStr = jsonStr.replace('[' + a + ']', '"' + strings[a] + '"');
    }

    jsonStr = jsonStr.replace(/</g, '[').replace(/>/g, ']');


    // get rid of duplicate quotes
    jsonStr = jsonStr.replace(/\"\"/g, '"');

    return jsonStr;
};

// this method takes in an object expression (e.g. object.property) and resolves it
var resolveObjectExpression = function (expression, context) {
    var objectName = 'window', object = context || window, property = null, propertyIsIndex = false;
    var objectFragments = expression.toString().split('.');

    for (var b = 0; b < objectFragments.length; b++) {
        property = objectFragments[b].trim();
        var tmp = null;
        var leftIdx = property.indexOf('['), rightIdx = property.indexOf(']');
        // handle special array syntax
        if (property.indexOf('[@each]') > -1) {
            objectName = property;
            object = object[property.replace('[@each]', '')];
            if (!object.computedProperties)
                object.computedProperties = [];
        }
        else if (property.indexOf('[i]') == 0) {

        }
        // handle array
        else if (leftIdx > -1 && rightIdx > -1) {
            var array = property.substring(0, leftIdx);
            var index = property.substring(leftIdx + 1, rightIdx);
            tmp = object[array][index];
        }
        else {
            tmp = object[property];
        }

        if (Lava.typeOf(tmp) === 'object') {
            objectName = property;
            object = tmp;
        }
    }

    // check to see if property is unique.  If not, we do not actually have a property in the expression, just an object
    if (objectName === property) {
        property = null;
    }

    // check to see if the property is actually an array index
    if (property && property.indexOf('[') > -1 && property.slice(-1) == ']') {
        objectName = property.substring(0, property.indexOf('['));
        object = object[objectName];
        property = property.substring(property.indexOf('[') + 1, property.length - 1);
        propertyIsIndex = true;
    }

    return { objectName: objectName, object: object, property: property, propertyIsIndex: propertyIsIndex };
};

var resolveAttribute = function (attribute) {
    if (attributeMaps[attribute])
        return attributeMaps[attribute];
    else
        return attribute;
};

var getPathFromPropertyExpression = function (expression) {
    return expression.substring(0, expression.lastIndexOf('.'));
};

var replaceTemplateTags = function (text) {
    var regex = new RegExp(Lava.dataTemplateName, 'g');
    return text.replace(regex, Lava.dataBindName);
};

var resolveAllUnboundExpressions = function (text) {
    var left = text.indexOf('{{'), right = 0;
    var replaces = [];

    while (left > -1) {
        // get the closing brace
        right = text.indexOf('}}', left);

        var expression = text.substring(left + 2, right);
        var result = resolveObjectExpression(expression);
        
        // replace the expression with its value
        var regex = new RegExp('\\{\\{' + expression.replace('[', '\\[').replace(']', '\\]').replace('(', '\\(').replace(')', '\\)') + '\\}\\}', 'g');
        var value = result.object[result.property];
        if (Lava.typeOf(result.object[result.property]) === 'function') {
            value = result.object[result.property]();
        }
        else if (Lava.typeOf(result.object[result.property]) === 'computed') {
            value = result.object[result.property].get();
        }
        

        replaces.push({ regex: regex, value: value });

        // get new index value
        left = text.indexOf('{{', right);
        var i = 0;
    }

    for (var i = 0; i < replaces.length; i++) {
        var replace = replaces[i];
        text = text.replace(replace.regex, replace.value);
    }

    return text;
}
